#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/wait.h>

#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"


#include "gpc_queue.h"
#include "gpc_support.h"
#include "gpc_autodetect.h"

#include "../../printerconf/src/printerconf.h"
#include "../pixmaps/gpc_banner.xpm"

#define MAX_PAGES 7

int pagenum = 0;

enum {
  WIZPG_START = 0,
  WIZPG_AUTO_OR_MANUAL,
  WIZPG_AUTO_INFO,
  WIZPG_AUTO_FOUND,
  WIZPG_SELECT_PRINTER_TYPE,
  WIZPG_SELECT_CONNECTION_TYPE,
  WIZPG_SET_DEFAULTS,
  WIZPG_LAST
} ENUM_WIZPG;

const char *prntype_names[] = 
{
  "Local",
  "Remote",
  "Direct-to-Port",
  "SMB (Samba)",
  "NCP (Netware)",
  "Unknown",
};

static GPC_PDINFO **pdinfo_list = NULL;

static char *alloc_str_glib2norm(char *str, int do_free)
{
  size_t len = strlen(str);
  char *ret = malloc(len + 1);

  if (!ret)
    abort();

  memcpy(ret, str, len + 1);

  if (do_free)
    g_free(str);

  return (ret);
}    

static char **alloc_strv_glib2norm(char **str, int do_free)
{
  size_t vec_len = 0;
  char **ret = NULL;
  
  while (str[vec_len])
    ++vec_len;

  ret = malloc(sizeof(char **) * (vec_len + 1));
  if (!ret)
    abort();

  vec_len = 0;
  while (str[vec_len])
  {
    ret[vec_len] = alloc_str_glib2norm(str[vec_len], FALSE);
    ++vec_len;
  }
  ret[vec_len] = NULL;

  if (do_free)
    g_strfreev(str);

  return (ret);
}

static void free_strv(char **str)
{
  unsigned int scan = 0;
  
  if (!str)
    return;

  while (str[scan])
    free(str[++scan]);
  free(str);
}

static void error_have_printer_name(const char *name)
{
  GString *str = g_string_new("");
  
  g_string_sprintf(str, _(" You already have a printer specified with "
                          "the name -- %s --."), name);
  gnome_warning_dialog(str->str);
  
  g_string_free(str, TRUE);
}

/*---------------------------------------------------------------
----------------------------------------------------------------*/
static int get_queue_info(GtkWidget * widget, GPC_QUEUE_INFO *qinfo,
                          int is_editing)
{
  GtkWidget *tmp = NULL;
  char *stmp = NULL;
  
  /* Get common info */
  tmp = lookup_widget(GTK_WIDGET(widget), "txtQueueName");

  stmp = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
  qinfo->qlist = alloc_strv_glib2norm(g_strsplit(stmp, "|", 0), TRUE);
  g_free(stmp);

  if (!is_editing)
  {
    unsigned int scan = 0;
    
    while (qinfo->qlist[scan])
    {
      if (gpc_queue_get_info(qinfo->qlist[scan]))
      {
        error_have_printer_name(qinfo->qlist[scan]);
        free_strv(qinfo->qlist);
        qinfo->qlist = NULL;
        
        return (FALSE);
      }
      
      ++scan;
    }
  }
  
  tmp = lookup_widget(GTK_WIDGET(widget), "txtSpoolDir");
  qinfo->spool_dir = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
  
  tmp = lookup_widget(GTK_WIDGET(widget), "txtFileLimit");
  qinfo->file_limit = (atoi(gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1)) *
                       1024);

  tmp = lookup_widget(GTK_WIDGET(widget), "chkSuppressHeaders");
  qinfo->suppress_headers = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp));

  return (TRUE);
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
void on_btnPrev_clicked(GtkButton *button,
			gpointer user_data __attribute((unused)))
{
  GtkWidget *notebook = lookup_widget (GTK_WIDGET (button), "nbkWizard");
  GtkWidget *window1 = lookup_widget (GTK_WIDGET (button), "window1");
  GtkWidget *widget;

  if (pagenum == WIZPG_SET_DEFAULTS) {
    gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFinish"));
    gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFilter"));
    gtk_widget_show(lookup_widget(GTK_WIDGET(button), "btnNext"));
  }

  /* Adjust page number. */
  switch (pagenum) {
  case WIZPG_SELECT_PRINTER_TYPE:
    pagenum = WIZPG_AUTO_OR_MANUAL;
    break;
  /*---------------------------------------------------------------*/
  case WIZPG_SELECT_CONNECTION_TYPE:
    /* Check which radiobutton was selected, move to
       appropriate next page. */
    widget = lookup_widget (GTK_WIDGET (button), "radManual");
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
      pagenum = WIZPG_SELECT_PRINTER_TYPE;
    else
      pagenum = WIZPG_AUTO_FOUND;
    break;
  /*---------------------------------------------------------------*/
  case WIZPG_SET_DEFAULTS:
    /* Check which radiobutton was selected, move to
       appropriate next page. */
    widget = lookup_widget (GTK_WIDGET (button), "radManual");
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
      pagenum = WIZPG_SELECT_CONNECTION_TYPE;
    else
      pagenum = WIZPG_AUTO_FOUND;
    break;
  /*---------------------------------------------------------------*/
  default:
    pagenum--;
  }

  if ((pagenum == WIZPG_START) ||
      (GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(window1),
                                           "is editing")) &&
       (pagenum == WIZPG_SELECT_PRINTER_TYPE)))
  {
    GtkWidget *tmp = lookup_widget(GTK_WIDGET(button), "btnPrev");
    gtk_widget_set_sensitive(tmp, FALSE);
  }
  
  /* Go to previous page. */
  if (pagenum < 0) 
    pagenum = 0;
  else
    gtk_notebook_set_page (GTK_NOTEBOOK (notebook), pagenum);

  /* Only make Next button work if both manufacturer and model are selected */
  if (pagenum == WIZPG_SELECT_PRINTER_TYPE) {
    gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (button), "btnNext"),
			      manuf_model_selected (GTK_WIDGET (button)));
  }
  else 
    gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (button), "btnNext"),
			      TRUE);

}

#define GPC__SET_TEXT(x, y) do { \
        int pos; \
        GtkWidget *tmp = lookup_widget(window1, x); \
        gtk_editable_delete_text(GTK_EDITABLE(tmp), 0, -1); \
        gtk_editable_insert_text(GTK_EDITABLE(tmp), \
                                  y, \
                                  strlen(y), &pos); \
} while (FALSE)

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
void on_btnNext_clicked(GtkButton *button,
			gpointer user_data __attribute((unused)))
{
  GtkWidget *notebook = lookup_widget (GTK_WIDGET (button), "nbkWizard");
  GtkWidget *widget;
  GtkWidget *window1 = lookup_widget(GTK_WIDGET(button), "window1");
  int is_editing = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(window1),
                                                       "is editing"));
  GPC_PDINFO *pdinfo = NULL;
  int row = -1;
  char *text = NULL;
  
  /* Get pertinent values from current page, adjust page number. */
  switch (pagenum)
  {  
    case WIZPG_AUTO_OR_MANUAL:
      /* Check which radiobutton was selected, move to
         appropriate next page. */
      widget = lookup_widget (GTK_WIDGET (button), "radManual");
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) 
        pagenum = WIZPG_SELECT_PRINTER_TYPE;
      else
        pagenum = WIZPG_AUTO_INFO;
      break;
      
    case WIZPG_AUTO_INFO:
    {
      char *in[4] = {NULL};
      char **out = NULL;
      int mask = 0;
      unsigned int scan = 0;
      int done = FALSE;
      
      widget = lookup_widget(GTK_WIDGET(button), "chkAutodetDirect");
      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
      {
        mask |= GPC_AUTODETECT_MASK_PARPORT;
        widget = lookup_widget(GTK_WIDGET(button), "txtDetectPort");
        in[GPC_AUTODETECT_OFF_PARPORT] = gtk_editable_get_chars(GTK_EDITABLE(widget),
                                                                0, -1);
      }
      
      widget = lookup_widget(GTK_WIDGET(button), "chkAutodetNetwork");
      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
      {
        mask |= GPC_AUTODETECT_MASK_NETWORK;
        widget = lookup_widget(GTK_WIDGET(button), "txtDetectIP");
        in[GPC_AUTODETECT_OFF_NETWORK] = gtk_editable_get_chars(GTK_EDITABLE(widget),
                                                                0, -1);
      }

      widget = lookup_widget(GTK_WIDGET(button), "chkAutodetUSB");
      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
      {
        mask |= GPC_AUTODETECT_MASK_NETWORK;
        widget = lookup_widget(GTK_WIDGET(button), "txtDetectUSB");
        in[GPC_AUTODETECT_OFF_NETWORK] = gtk_editable_get_chars(GTK_EDITABLE(widget),
                                                                0, -1);
      }
      
      out = gpc_autodetect(mask, (const char **)in);

      scan = 0;
      while (scan < 4)
      {
        g_free(in[scan]);
        ++scan;
      }

      scan = 0;
      if (out)
        while (out[scan])
        {
          char **a = g_strsplit(out[scan], ";", 0);
          char *ins[3];
          unsigned int a_ents = 0;
          unsigned int text_scan = 0;
          char *new_model = NULL;
          const char *vendor = NULL;
          
          while (a[a_ents])
            ++a_ents;
          
          text_scan = 0;
          while (text_scan < a_ents)
          {
            if (!strncmp("model=", a[text_scan], strlen("model=")))
            {
              ins[0] = a[text_scan] + strlen("model=");
            }
            else if (!strncmp("printer=", a[text_scan], strlen("printer=")))
            {
              ins[1] = a[text_scan];
            }
            else if (!strncmp("port=", a[text_scan], strlen("port=")))
            {
              ins[1] = a[text_scan];
            }
            else if (!strncmp("vendor=", a[text_scan], strlen("vendor=")))
            {
              vendor = a[text_scan] + strlen("vendor=");
            }
            else
            {
              fprintf(stderr, " Unknown gpc_autodetect() str -- %s --.\n",
                      out[scan]);
              goto next_loop;
            }
          }

          if (!ins[0] || !ins[1])
          {
            fprintf(stderr, " Missing gpc_autodetect() str -- %s --.\n",
                    out[scan]);
            goto next_loop;
          }

          if (ins[1] && !strncmp("port=", ins[1], strlen("port=")) &&
              !strcmp(ins[0], "Port not configured"))
            goto next_loop;

          pdinfo = gpc_pdinfo_find(pdinfo_list, ins[0]);
          if (!pdinfo && vendor)
          {
            GString *str = g_string_new("");

            g_string_sprintf(str, "%s %s", vendor, ins[0]);
            new_model = str->str;

            g_string_free(str, FALSE);

            pdinfo = gpc_pdinfo_find(pdinfo_list, new_model);
          }
          
          if (!pdinfo)
          {
            GString *str = g_string_new("");

            g_string_sprintf(str, " A printer model -- %s -- was"
                             " autodetected but is not present in the gpc"
                             " printer database, so it can't be used.\n",
                             ins[0]);
            gnome_warning_dialog(str->str);

            g_string_free(str, TRUE);
            goto next_loop;
          }
          ins[0] = pdinfo->brief_desc;
          
          widget = lookup_widget(GTK_WIDGET(button), "clist1");
          
          if (!done)
          {
            done = TRUE;
            gtk_clist_clear(GTK_CLIST(widget));
          }

          ins[2] = NULL;
          gtk_clist_append(GTK_CLIST(widget), ins);
          
          done = TRUE;
          
         next_loop:
          g_strfreev(a);
          free(out[scan]);
          g_free(new_model);
          ++scan;
        }
      free(out);
      
      if (done)
        ++pagenum;
      else
      {
        gnome_warning_dialog("Couldn't autodetect any printers.\n");
        return;
      }
    }
    break;

    case WIZPG_AUTO_FOUND:
      widget = lookup_widget(GTK_WIDGET(button), "clist1");
      row = GPOINTER_TO_INT(GTK_CLIST(widget)->selection->data);
      if (row >= 0)
        gtk_clist_get_text(GTK_CLIST(widget), row, 0, (gchar **) &text);

      pdinfo = gpc_pdinfo_description_find(pdinfo_list, text);
      if (!pdinfo)
      {
        GString *str = g_string_new("");
        
        g_string_sprintf(str, " The printer -- %s -- wasn't"
                         " present in the gpc printer database, however it had"
                         " to be there previously so I'm confused.\n",
                         text);
        gnome_warning_dialog(str->str);
        
        g_string_free(str, TRUE);
        return;
      }
      
      row = GPOINTER_TO_INT(GTK_CLIST(widget)->selection->data);
      if (row >= 0)
        gtk_clist_get_text(GTK_CLIST(widget), row, 1, (gchar **) &text);

      if (!strncmp("port=", text, strlen("port=")))
      {
        widget = lookup_widget(GTK_WIDGET(button), "radConnLocal");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), TRUE);
        pagenum = WIZPG_SET_DEFAULTS;
      }
      else if (!strncmp("printer=", text, strlen("printer=")))
      {
        widget = lookup_widget(GTK_WIDGET(button), "radConnRemote");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), TRUE);
        pagenum = WIZPG_SELECT_CONNECTION_TYPE;
        
        widget = lookup_widget(GTK_WIDGET(button), "txtRemoteHost");
        gtk_entry_set_text(GTK_ENTRY(widget), text + strlen("printer="));
        widget = lookup_widget(GTK_WIDGET(button), "txtDirectHost");
        gtk_entry_set_text(GTK_ENTRY(widget), text + strlen("printer="));
      }
      else
      {
        gnome_warning_dialog("Couldn't work out the type of the printer.\n");
        return;
      }
      break;

    case WIZPG_SELECT_CONNECTION_TYPE:
      widget = lookup_widget(GTK_WIDGET(button), "cltPrinterModel");
      text = clist_get_first_selected_text(GTK_CLIST(widget));
      pdinfo = gpc_pdinfo_description_find(pdinfo_list, text);
      
      if (!pdinfo)
      {
        gnome_warning_dialog("Couldn't find the name of the printer.\n");
        return;
      }
      pagenum = WIZPG_SET_DEFAULTS;
      break;
      
    default:
      ++pagenum;
  }

  if (pagenum != WIZPG_START)
  { /* works for is editing */
    GtkWidget *tmp = lookup_widget(GTK_WIDGET(button), "btnPrev");
    gtk_widget_set_sensitive(tmp, TRUE);
  }

  /* Go to next page. */
  if (pagenum > MAX_PAGES-1) 
    pagenum = MAX_PAGES-1;

  gtk_notebook_set_page (GTK_NOTEBOOK (notebook), pagenum);

  switch (pagenum)
  {
    case WIZPG_SELECT_PRINTER_TYPE:
    gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (button), "btnNext"),
			      manuf_model_selected (GTK_WIDGET (button)));
    break;
    
    case WIZPG_SET_DEFAULTS:
    {
      unsigned int scan = 0;
      GString *str = g_string_new("lp");

      g_assert(pdinfo);
      
      gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnNext"));
      gtk_widget_show(lookup_widget(GTK_WIDGET(button), "btnFinish"));
      gtk_widget_show(lookup_widget(GTK_WIDGET(button), "btnFilter"));
      
      gtk_notebook_set_show_tabs 
        (GTK_NOTEBOOK (lookup_widget (GTK_WIDGET (button), "nbkSetDefs")), 
         FALSE);
      
      gtk_label_set_text
        (GTK_LABEL (lookup_widget (GTK_WIDGET (button), "lblSetDefsPname")),
         pdinfo ? pdinfo->brief_desc : "** ERROR **");
      gtk_object_set_data(GTK_OBJECT(window1), "pdinfo entry", pdinfo);
      
      if (!is_editing)
      {
        /* in theory if we have UINT_MAX printers, like yeh right */
        while (gpc_queue_get_info(str->str))
          g_string_sprintf(str, "lp%u", ++scan);
        
        GPC__SET_TEXT("txtQueueName", str->str);
        
        g_string_prepend(str, "/var/spool/lpd/");
        GPC__SET_TEXT("txtSpoolDir", str->str);
      }
      
      g_string_free(str, TRUE);
    }
    break;

    case WIZPG_AUTO_FOUND:
      gtk_widget_set_sensitive(lookup_widget(GTK_WIDGET(button), "btnNext"),
                               auto_model_selected(GTK_WIDGET(button)));
      break;
      
    default:
      gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (button), "btnNext"),
                                TRUE);
      break;
  }
}
#undef GPC__SET_TEXT

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
void on_btnCancel_clicked(GtkButton *button,
			  gpointer user_data __attribute((unused)))
{
  GtkWidget *tmp = NULL;
  
  gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "window1"));
  gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFinish"));
  gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFilter"));
  tmp = lookup_widget(GTK_WIDGET(button), "btnNext");
  gtk_widget_show(tmp);
  gtk_widget_set_sensitive(tmp, TRUE);
  tmp = lookup_widget(GTK_WIDGET(button), "btnPrev");
  gtk_widget_set_sensitive(tmp, FALSE);
  tmp = lookup_widget(GTK_WIDGET(button), "txtQueueName");
  gtk_widget_set_sensitive(tmp, TRUE);
  tmp = lookup_widget(GTK_WIDGET(button), "nbkWizard");
  gtk_notebook_set_page(GTK_NOTEBOOK(tmp), pagenum = WIZPG_START);
}

/*----------------------------------------------------------------
  When a manufacturer is selected from the clist widget.
----------------------------------------------------------------*/
void on_cltManuf_select_row(GtkCList *clist,gint row,gint column,
			    GdkEvent *event __attribute((unused)),
			    gpointer user_data __attribute((unused)))
{
  gchar *manuf_name;
  GPC_PDINFO ** manuf_list = NULL;

  GtkWidget *cltPrinterModel = lookup_widget 
    (GTK_WIDGET (clist), "cltPrinterModel");

  if (gtk_clist_get_text (clist, row, column, (gchar **) &manuf_name)) {
    gtk_clist_clear (GTK_CLIST (cltPrinterModel));

    manuf_list = gpc_pdinfo_get_manuf_list (pdinfo_list, manuf_name);
    if (manuf_list)
      pdinfo_to_clist (manuf_list, cltPrinterModel);

  }
}


/*-------------------------------------------------------------------
  This handles behavior if any of the radConn... radio buttons is
  toggled.
-------------------------------------------------------------------*/
void on_radConn_toggled(GtkToggleButton *togglebutton,
			gpointer user_data __attribute((unused)))
{
  int page = -1;
  GtkWidget *nbkSetDefs = NULL;
  char *title = NULL;
  const char *format = "%s Printer Setup";

  if (gtk_toggle_button_get_active (togglebutton)) {
    const int page_num2enum_map[] = 
      { /* mapping from the order layed out in the notebook to the enum */
       GPC_QTYPE_LOCAL,
       GPC_QTYPE_REMOTE,
       GPC_QTYPE_SMB,
       GPC_QTYPE_NCP,
       GPC_QTYPE_DIRECT,
      };
    
    page = (int) gtk_object_get_data (GTK_OBJECT (togglebutton), NP);
    nbkSetDefs = lookup_widget(GTK_WIDGET (togglebutton), "nbkSetDefs");
    gtk_notebook_set_page (GTK_NOTEBOOK (nbkSetDefs), page);

    title = g_malloc(strlen(format) +
                     strlen(prntype_names[page_num2enum_map[page]]) + 1);

    sprintf (title, format, prntype_names[page_num2enum_map[page]]);

    gtk_label_set_text 
      (GTK_LABEL (lookup_widget (GTK_WIDGET (togglebutton), 
				 "lblSetDefsHdr")), title);

    gtk_object_set_data 
      (GTK_OBJECT (lookup_widget (GTK_WIDGET (togglebutton), "lblSetDefsHdr")), 
       CT, (gpointer) page_num2enum_map[page]);

    g_free (title);
  }

}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_cltPrinterModel_select_row(GtkCList *clist,
				   gint row __attribute((unused)),
				   gint column __attribute((unused)),
				   GdkEvent *event __attribute((unused)),
				   gpointer user_data __attribute((unused)))
{
  if (pagenum == WIZPG_SELECT_PRINTER_TYPE)
    gtk_widget_set_sensitive 
      (GTK_WIDGET (lookup_widget (GTK_WIDGET (clist), "btnNext")),
       manuf_model_selected (GTK_WIDGET (clist)));

}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_cltManuf_select_row_set_btnNext(GtkCList *clist,
					gint row __attribute((unused)),
					gint column __attribute((unused)),
					GdkEvent *event __attribute((unused)),
					gpointer user_data __attribute((unused)))
{
  if (pagenum == WIZPG_SELECT_PRINTER_TYPE)
    gtk_widget_set_sensitive 
      (GTK_WIDGET (lookup_widget (GTK_WIDGET (clist), "btnNext")),
       manuf_model_selected (GTK_WIDGET (clist)));

}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
static int entry_is_blank(GtkWidget *entry)
{
  gchar *str = NULL;

  str = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);  

  str = chop (str);

  return (str[0] == '\0');

}

/*----------------------------------------------------------------
  The type should be a valid gnome message box type 
  (GNOME_MESSAGE_BOX_ERROR, etc.).
----------------------------------------------------------------*/
static void msgbox(const char *type, const char *message)
{
  GtkWidget *infobox = gnome_message_box_new (message, type,
                                              GNOME_STOCK_BUTTON_OK, NULL);

  gnome_dialog_run (GNOME_DIALOG (infobox));
}

static void create_spool_dir_yes_no(int reply, void *data)
{
  GPC_QUEUE_INFO *qinfo = NULL;
  GString *str = g_string_new("");
  
  if (!(qinfo = gpc_queue_get_info((char *)data)))
  {
    g_string_sprintf(str, " The printer -- %s -- doesn't exist anymore.",
                     (char *)data);
    gnome_warning_dialog(str->str);
    goto cleanup;
  }
  
  switch (reply)
  {
   case GNOME_YES:
   {
     char *bad_path = _pconf_mkdir_p(qinfo->spool_dir, 0755);
     struct passwd *pw = NULL;
     struct group *gr = NULL;
     
     if (bad_path)
     {
       g_string_sprintf(str, " While trying to create the spool directory"
                        " -- %s -- for the printer -- %s -- the error"
                        " -- %s -- was encountered on the path -- %s --.",
                        qinfo->spool_dir, qinfo->qlist[0], strerror(errno),
                        bad_path);
       g_free(bad_path);
       gnome_warning_dialog(str->str);
       goto cleanup;
     }

     if (!(pw = getpwnam("root")))
     {
       g_string_sprintf(str, " Although the spool directory"
                        " -- %s -- exists the username -- root -- could not be"
                        " looked up in the password database. This is needed"
                        " to change the ownership of the directory.",
                        qinfo->spool_dir);
       gnome_warning_dialog(str->str);
       goto cleanup;
     }

     if (!(gr = getgrnam("lp")))
     {
       g_string_sprintf(str, " Although the spool directory"
                        " -- %s -- exists the groupname -- lp -- could not be"
                        " looked up in the group database. This is needed"
                        " to change the ownership of the directory.",
                        qinfo->spool_dir);
       gnome_warning_dialog(str->str);
       goto cleanup;
     }

     
     if (chown(qinfo->spool_dir, pw->pw_uid, gr->gr_gid) == -1)
     {
       g_string_sprintf(str, " While trying to change the ownership of the"
                        " spool directory"
                        " -- %s -- for the printer -- %s -- the error"
                        " -- %s -- was encountered.",
                        qinfo->spool_dir, qinfo->qlist[0], strerror(errno));
       gnome_warning_dialog(str->str);
       goto cleanup;
     }

     if (chmod(qinfo->spool_dir, 0775) == -1)
     {
       g_string_sprintf(str, " While trying to change the permissions of the"
                        " spool directory"
                        " -- %s -- for the printer -- %s -- the error"
                        " -- %s -- was encountered.",
                        qinfo->spool_dir, qinfo->qlist[0], strerror(errno));
       gnome_warning_dialog(str->str);
       goto cleanup;
     }  
   }
   break;
   
    case GNOME_NO:
      break;
      
    default:
      g_assert_not_reached();
  }
  
 cleanup:
  g_string_free(str, TRUE);

  g_free(data);
}


#define MV_GSTRING(ptr, str) do { \
  (ptr) = (str)->str; \
  g_string_free((str), FALSE); \
  (str) = g_string_new(""); \
  } while (FALSE)

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_btnFinish_clicked(GtkButton *button,
			  gpointer user_data __attribute((unused)))
{
  GPC_QUEUE_INFO *qinfo = NULL;
  GtkWidget *tmp = NULL;
  int ok = TRUE;
  GtkWidget *window1 = lookup_widget(GTK_WIDGET(button), "window1");
  GtkWidget *wndMain = gtk_object_get_data(GTK_OBJECT(window1), "main");
  int is_editing = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(window1),
                                                       "is editing"));
  GString *str = g_string_new("");
  GPC_PDINFO *pdinfo = NULL;
  GPC_FILTER_SELECTION *filter = NULL;
  
  /* Validate common data (top half of window). */
  if (entry_is_blank (lookup_widget (GTK_WIDGET (button), "txtQueueName"))) {
    msgbox (GNOME_MESSAGE_BOX_ERROR, "Queue Name cannot be blank");
    g_string_free(str, TRUE);
    return;
  }

  qinfo = g_malloc0(sizeof(GPC_QUEUE_INFO));

  pdinfo = gtk_object_get_data(GTK_OBJECT(window1), "pdinfo entry");
  g_assert(pdinfo);

  qinfo->filter_opts = gtk_object_get_data(GTK_OBJECT(window1),
                                           "device filter options");
  if (!qinfo->filter_opts)
  {
    filter = g_malloc0(sizeof(GPC_FILTER_SELECTION));
    qinfo->filter_opts = filter;
  }
  
  if (!qinfo->filter_opts->res && pdinfo->res_list && pdinfo->res_list[0])
  {
    g_string_sprintf(str, "%u%s%u", pdinfo->res_list[0]->x, "x",
                     pdinfo->res_list[0]->y);
    MV_GSTRING(qinfo->filter_opts->res, str);
  }

  if (!qinfo->filter_opts->size)
    qinfo->filter_opts->size = g_strdup("letter"); /* FIXME: US -> uk needs to be a4 */
    
  if (!qinfo->filter_opts->bpp && pdinfo->bits_list && pdinfo->bits_list[0])
  {
    g_string_sprintf(str, "%s", pdinfo->bits_list[0]->bitsperpixel);
    MV_GSTRING(qinfo->filter_opts->bpp, str);
  }

  if (!qinfo->filter_opts->pages_per_page)
    qinfo->filter_opts->pages_per_page = g_strdup("1");

  if (!qinfo->filter_opts->printer_model)
  {
    g_assert(filter); /* should only happen if there was no filter at all */
    qinfo->filter_opts->printer_model = g_strdup(pdinfo->brief_desc);
  }
  
  if (!qinfo->filter_opts->gs_dev)
  {
    g_assert(filter); /* should only happen if there was no filter at all */
    qinfo->filter_opts->gs_dev = g_strdup(pdinfo->gsdriver);
  }
  
  tmp = lookup_widget(GTK_WIDGET (button), "lblSetDefsHdr");
  qinfo->qtype = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(tmp), CT));

  ok = get_queue_info(GTK_WIDGET(button), qinfo, is_editing);
  if (!ok)
  {
    g_free(qinfo);
    g_string_free(str, TRUE);
    return;
  }

  if (!g_file_test(qinfo->spool_dir, G_FILE_TEST_ISDIR))
  {
    g_string_sprintf(str, " The spool directory -- %s -- for the printer"
                     " -- %s -- does not exist. Do you want GPC to create this"
                     " now?\n Note that your printer will not work until it is"
                     " created.",
                     qinfo->spool_dir, qinfo->qlist[0]);

    gnome_question_dialog(str->str, create_spool_dir_yes_no,
                          g_strdup(qinfo->qlist[0]));
  }

  switch (qinfo->qtype)
  {
    case GPC_QTYPE_LOCAL:
      tmp = lookup_widget(GTK_WIDGET(button), "txtLocalDevice");
      qinfo->local_device = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      if (!g_file_test(qinfo->local_device, G_FILE_TEST_ISFILE))
      {
        g_string_sprintf(str, " The local device -- %s -- for the printer"
                         " -- %s -- does not exist, you will need to create"
                         " this before your printer will work.",
                         qinfo->local_device, qinfo->qlist[0]);
        gnome_warning_dialog(str->str);
      }
      break;
    case GPC_QTYPE_REMOTE:
      tmp = lookup_widget(GTK_WIDGET(button), "txtRemoteHost");
      qinfo->remote_host = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);

      tmp = lookup_widget(GTK_WIDGET(button), "txtRemoteQueue");
      qinfo->remote_queue = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      break;
    case GPC_QTYPE_DIRECT:
      tmp = lookup_widget(GTK_WIDGET(button), "txtDirectHost");
      qinfo->direct_host = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);

      tmp = lookup_widget(GTK_WIDGET(button), "txtDirectPort");
      qinfo->direct_port = atoi(gtk_editable_get_chars(GTK_EDITABLE(tmp),
                                                       0, -1));
      break;
    case GPC_QTYPE_SMB:
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBHost");
      qinfo->SMB_host = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBIP");
      qinfo->SMB_ip = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBUser");
      qinfo->user = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBPassword");
      qinfo->passwd = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBPName");
      qinfo->SMB_printer_name = gtk_editable_get_chars(GTK_EDITABLE(tmp),
                                                       0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtSMBGroup");
      qinfo->SMB_workgroup = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      break;
    case GPC_QTYPE_NCP:
      tmp = lookup_widget(GTK_WIDGET(button), "txtNCPSName");
      qinfo->NCP_printserver_name = gtk_editable_get_chars(GTK_EDITABLE(tmp),
                                                           0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtNCPQname");
      qinfo->NCP_queue_name = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtNCPUser");
      qinfo->user = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      tmp = lookup_widget(GTK_WIDGET(button), "txtNCPPassword");
      qinfo->passwd = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1);
      break;
    default:
      g_assert_not_reached();
      break;
  }  

  if (is_editing)
  { /* FIXME: can't change names atm. -- this is sort of ok though as
     * you can alias to another and delete the old name */
    gpc_queue_edt(gpc_queue_get_list(NULL), qinfo);
  }
  else
  {
    unsigned int scan = 0;
    GtkWidget *clist = NULL;
    
    gpc_queue_add(gpc_queue_get_list(NULL), qinfo);
    
    clist = lookup_widget(wndMain, "clist2");

    while (qinfo->qlist[scan])
    {
      char *a[2];
      
      a[0] = qinfo->qlist[scan];
      a[1] = NULL;
      
      gtk_clist_append(GTK_CLIST(clist), a);
    
      ++scan;
    }
  }
  gtk_clist_unselect_all(GTK_CLIST(lookup_widget(wndMain, "clist2")));
                         
  gtk_widget_hide(window1);
  gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFinish"));
  gtk_widget_hide(lookup_widget(GTK_WIDGET(button), "btnFilter"));
  tmp = lookup_widget(GTK_WIDGET(button), "btnNext");
  gtk_widget_show(tmp);
  gtk_widget_set_sensitive(tmp, TRUE);
  tmp = lookup_widget(GTK_WIDGET(button), "btnPrev");
  gtk_widget_set_sensitive(tmp, FALSE);
  tmp = lookup_widget(GTK_WIDGET(button), "txtQueueName");
  gtk_widget_set_sensitive(tmp, TRUE);
  tmp = lookup_widget(GTK_WIDGET(button), "nbkWizard");
  gtk_notebook_set_page(GTK_NOTEBOOK(tmp), pagenum = WIZPG_START);

  g_string_free(str, TRUE);
}
#undef MV_GSTRING

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_txtLocalDevice_changed (GtkEditable *editable __attribute((unused)), 
				gpointer user_data __attribute((unused)))
{
#if 0
  int conntype = -1;
  GPC_QUEUE_INFO *qinfo = NULL;

  conntype = (int) gtk_object_get_data 
    (GTK_OBJECT(lookup_widget (GTK_WIDGET (editable), "lblSetDefsHdr")), CT);
  printf ("-->%s\n", prntype_names[conntype]);

  /* qinfo = (GPC_QUEUE_INFO *) malloc (sizeof (GPC_QUEUE_INFO));
     get_queue_info (GTK_WIDGET (editable), conntype, qinfo);
     
     printf ("-->%s\n", qinfo->qname);  */
  g_warning ("This is a pop-up");

  /*free_queue_info_fields (qinfo);*/
  free (qinfo);
#endif
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_exit1_activate(GtkMenuItem *menuitem __attribute((unused)),
		       gpointer user_data __attribute((unused))){
  gtk_main_quit();
}


void on_wndMain_destroy(GtkObject *object __attribute((unused)),
			gpointer user_data __attribute((unused))){
  gtk_main_quit();
}


/*----------------------------------------------------------------
----------------------------------------------------------------*/
static int
add_printer_beg(GtkWidget *window1)
{
  GtkWidget *clist;
  GtkWidget *tmp = NULL;
  char **manuf_list = NULL;
  static int first_time = 1;

  gtk_window_set_title(GTK_WINDOW(window1), _("GPC - Add Printer"));
  
  if (first_time == 1)
  {
    /* could do a pdinfo_free_list() when hiding the window again */
    if (g_file_exists(GPC_SHARE_PREFIX "/printerdb.xml"))
      pdinfo_list = gpc_pdinfo_get_list (GPC_SHARE_PREFIX "printerdb.xml", NULL);
    else if (g_file_exists("./src/printerdb.xml"))
      pdinfo_list = gpc_pdinfo_get_list ("./src/printerdb.xml", NULL);
    else if (g_file_exists("./printerdb.xml"))
      pdinfo_list = gpc_pdinfo_get_list ("./printerdb.xml", NULL);

    if (!pdinfo_list)
    {
      gnome_warning_dialog(" The data file for GPC was searched for in...\n"
                           " 1. " GPC_SHARE_PREFIX "/printerdb.xml" "\n"
                           " 2. " "./src/printerdb.xml" "\n"
                           " 3. " "./printerdb.xml" "\n"
                           "...but could not be found. This is needed.");
      return (FALSE);
    }
    
    manuf_list = gpc_pdinfo_get_manuf_names(pdinfo_list);
    
    /* Set data for cltManuf clist */
    clist = lookup_widget (window1, "cltManuf");
    
    
    /* make sure tabs aren't visible on main notebook widget */
    tmp = lookup_widget(window1, "nbkWizard");
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(tmp), FALSE);
    
    /* Set data for radConn... radiobuttons */
    tmp = lookup_widget(window1, "radConnLocal");
    gtk_object_set_data(GTK_OBJECT(tmp), NP, (gpointer) 0);
    tmp = lookup_widget(window1, "radConnRemote");
    gtk_object_set_data(GTK_OBJECT(tmp), NP, (gpointer) 1);
    tmp = lookup_widget(window1, "radConnSMB");
    gtk_object_set_data(GTK_OBJECT(tmp), NP, (gpointer) 2);
    tmp = lookup_widget(window1, "radConnNCP");
    gtk_object_set_data(GTK_OBJECT(tmp), NP, (gpointer) 3);
    tmp = lookup_widget(window1, "radConnDirect");
    gtk_object_set_data(GTK_OBJECT(tmp), NP, (gpointer) 4);
    
    charlist_to_clist(manuf_list, clist);
    gtk_clist_unselect_all(GTK_CLIST(clist));
    
    set_pixmap_xpm (lookup_widget (window1, "pixmap1"), gpc_banner);
    
    set_widgets_font (window1,
                      "-adobe-helvetica-bold-r-normal-*-*-140-*-*-*-*-*-*",
                      "label8", "lblSetDefsHdr",
                      NULL);
    
    on_radConn_toggled
      (GTK_TOGGLE_BUTTON (lookup_widget (GTK_WIDGET (window1), "radConnLocal")),
       NULL);
    
    first_time = 0;
  }
  
  gtk_object_set_data(GTK_OBJECT(window1), "is editing", (gpointer) FALSE);

  return (TRUE);
}

static void
del_printer_beg(GtkCList *clist)
{
  if (clist->selection)
  {
    int row = GPOINTER_TO_INT(clist->selection->data);
    char *text = NULL;

    if (row >= 0)
    {
      gtk_clist_get_text(clist, row, 0, &text);
      
      gpc_queue_del(gpc_queue_get_list(NULL), text);
      gtk_clist_remove(clist, row);
    }
  }
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
static int gpc__find_row_text(GtkCList *clist, const char *text)
{
  int scan = 0;
  
  while (scan <= clist->rows)
  {
    char *row_text = NULL;

    gtk_clist_get_text(clist, scan, 0, &row_text);

    g_assert(row_text);
    if (row_text && !strcmp(row_text, text))
      return (scan);
    
    ++scan;
  }
  
  return (-1);
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
#define GPC__SET_TEXT(x, y) do { \
        int pos = 0; \
        tmp = lookup_widget(window1, x); \
        gtk_editable_delete_text(GTK_EDITABLE(tmp), 0, -1); \
        gtk_editable_insert_text(GTK_EDITABLE(tmp), \
                                  y, \
                                  strlen(y), &pos); \
} while (FALSE)

#define GPC__Q_DUP(filter, old, x) do { \
 if ((old)->x) \
   (filter)->x = g_strdup((old)->x); \
 } while (FALSE)

static void
edt_printer_beg(GtkWidget *window1, GtkCList *clist)
{
  if (clist->selection)
  {
    int row = GPOINTER_TO_INT(clist->selection->data);
    char *text = NULL;
    GPC_QUEUE_INFO *qinfo = NULL;
    GString *str = g_string_new("");
    GtkWidget *tmp = NULL;
    char *qnames = NULL;
    GPC_PDINFO *pdinfo = NULL;
    
    gtk_clist_get_text(clist, row, 0, &text);

    if (row < 0)
      return;

    if (!add_printer_beg(window1))
      return;
    
    qinfo = gpc_queue_get_info(text);

    if (qinfo->filter_opts && qinfo->filter_opts->printer_model)
      text = qinfo->filter_opts->printer_model;
    g_string_sprintf(str, _("GPC - Editing Printer %s"), text);
    
    gtk_window_set_title(GTK_WINDOW(window1), str->str);

    gtk_object_set_data(GTK_OBJECT(window1), "is editing", (gpointer) TRUE);
    
    tmp = lookup_widget(window1, "nbkWizard");
    gtk_notebook_set_page(GTK_NOTEBOOK(tmp),
                          pagenum = WIZPG_SELECT_PRINTER_TYPE);

    if (qinfo->filter_opts)
    {
      GPC_FILTER_SELECTION *filter = g_malloc0(sizeof(GPC_FILTER_SELECTION));
      GPC_FILTER_SELECTION *old = qinfo->filter_opts;

      GPC__Q_DUP(filter, old, printer_model);
      GPC__Q_DUP(filter, old, gs_dev);
      GPC__Q_DUP(filter, old, res);
      GPC__Q_DUP(filter, old, size);
      GPC__Q_DUP(filter, old, bpp);
      GPC__Q_DUP(filter, old, margin_right_left);
      GPC__Q_DUP(filter, old, margin_top_bottom);
      GPC__Q_DUP(filter, old, pages_per_page);
      GPC__Q_DUP(filter, old, gs_opts);

      filter->send_eof = old->send_eof;
      filter->fix_stair_text = old->fix_stair_text;
      filter->fast_text = old->fast_text;
      
      gtk_object_set_data(GTK_OBJECT(window1), "device filter options",
                          (gpointer)filter);
    }
    
  if (qinfo->filter_opts && qinfo->filter_opts->printer_model && 
        (pdinfo = gpc_pdinfo_description_find(pdinfo_list,
                                              qinfo->filter_opts->printer_model)))
    {
      tmp = lookup_widget(window1, "cltManuf");
      row = gpc__find_row_text(GTK_CLIST(tmp), pdinfo->printer_manuf);
      gtk_clist_select_row(GTK_CLIST(tmp), row, 0);
      
      tmp = lookup_widget(window1, "cltPrinterModel");
      row = gpc__find_row_text(GTK_CLIST(tmp),
                               qinfo->filter_opts->printer_model);
      gtk_clist_select_row(GTK_CLIST(tmp), row, 0);
    }

    /* it crashes if this is wrong. */
    gtk_widget_set_sensitive(lookup_widget(window1, "btnNext"),
                             manuf_model_selected(window1));

    qnames = gpc_queue_names(qinfo);
    GPC__SET_TEXT("txtQueueName", qnames);
    g_free(qnames);
    gtk_widget_set_sensitive(tmp, FALSE); /* txtQueueName */

    tmp = lookup_widget(window1, "chkSuppressHeaders");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp),
                                 qinfo->suppress_headers);

    GPC__SET_TEXT("txtSpoolDir", qinfo->spool_dir); 
    
    g_string_sprintf(str, "%d", qinfo->file_limit / 1024);
    GPC__SET_TEXT("txtFileLimit", str->str);
    
    switch(qinfo->qtype)
    {
      case GPC_QTYPE_LOCAL:
        tmp = lookup_widget(window1, "radConnLocal");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
        GPC__SET_TEXT("txtLocalDevice", qinfo->local_device);
        break;
      case GPC_QTYPE_REMOTE:
        tmp = lookup_widget(window1, "radConnRemote");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
        GPC__SET_TEXT("txtRemoteHost", qinfo->remote_host);
        GPC__SET_TEXT("txtRemoteQueue", qinfo->remote_queue);
        break;
      case GPC_QTYPE_DIRECT:
        tmp = lookup_widget(window1, "radConnDirect");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
        GPC__SET_TEXT("txtDirectHost", qinfo->direct_host);
        g_string_sprintf(str, "%d", qinfo->direct_port);
        GPC__SET_TEXT("txtDirectPort", str->str);
        break;
      case GPC_QTYPE_SMB:
        tmp = lookup_widget(window1, "radConnSMB");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
        GPC__SET_TEXT("txtSMBHost", qinfo->SMB_host);
        GPC__SET_TEXT("txtSMBIP", qinfo->SMB_ip);
        GPC__SET_TEXT("txtSMBUser", qinfo->user);
        GPC__SET_TEXT("txtSMBPassword", qinfo->passwd);
        GPC__SET_TEXT("txtSMBPName", qinfo->SMB_printer_name);
        GPC__SET_TEXT("txtSMBGroup", qinfo->SMB_workgroup);
        break;
      case GPC_QTYPE_NCP:
        tmp = lookup_widget(window1, "radConnNCP");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
        GPC__SET_TEXT("txtNCPSName", qinfo->NCP_printserver_name);
        GPC__SET_TEXT("txtNCPQname", qinfo->NCP_queue_name);
        GPC__SET_TEXT("txtNCPUser", qinfo->user);
        GPC__SET_TEXT("txtNCPPassword", qinfo->passwd);
        break;
      default:
        g_assert_not_reached();
        break;
    }

    g_string_free(str, TRUE);
    
    gtk_widget_show(window1);
  }
}
#undef GPC__SET_TEXT
#undef GPC__Q_DUP

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_add1_activate(GtkMenuItem *menuitem,
		      gpointer user_data __attribute((unused)))
{
  GtkWidget *wndMain = lookup_widget(GTK_WIDGET(menuitem), "wndMain");
  GtkWidget *window1 = gtk_object_get_data(GTK_OBJECT(wndMain), AP);

  if (!add_printer_beg(window1))
    return;
  
  gtk_widget_show(window1);
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_remove1_activate(GtkMenuItem *menuitem,
			 gpointer user_data __attribute((unused)))
{
  GtkWidget *clist = lookup_widget(GTK_WIDGET(menuitem), "clist2");
  
  del_printer_beg(GTK_CLIST(clist));
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_btnAddPrinter_clicked(GtkButton *button __attribute((unused)),
			      gpointer user_data)
{
  GtkWidget *wndMain = user_data;
  GtkWidget *window1 = gtk_object_get_data(GTK_OBJECT(wndMain), AP);

  if (!add_printer_beg(window1))
    return;
  
  gtk_widget_show(window1);
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
void on_btnDelPrinter_clicked(GtkButton *button __attribute((unused)),
			      gpointer user_data)
{
  GtkCList *clist = user_data;
  del_printer_beg(clist);
}


void on_btnEditPrinter_clicked(GtkButton *button __attribute((unused)),
			       gpointer user_data)
{
  GtkWidget *wndMain = user_data;
  GtkWidget *clist = lookup_widget(wndMain, "clist2");
  GtkWidget *window1 = gtk_object_get_data(GTK_OBJECT(wndMain), AP);

  edt_printer_beg(window1, GTK_CLIST(clist));
}


void on_btnClose_clicked(GtkButton *button __attribute((unused)),
			 gpointer user_data __attribute((unused))){
  gtk_main_quit();
}


void on_about1_activate(GtkMenuItem *menuitem __attribute((unused)),
			gpointer user_data __attribute((unused))){
  GtkWidget *window = NULL;
  const char *authors[] = {
   N_("Mark Pruett."),
   N_("James Antill."),
   N_("Ben Woodard."),
   NULL
  };
  
  window = gnome_about_new(_("GPC"), "0.1 of GPC",
                           "(C) 2001 Hewlett Packard; Released under the GPL",
                           authors,
                           _("Gnome Print Controller."),
                           NULL);
  
  gtk_widget_show(window);
}

void on_clist2_select_row(GtkCList *clist,gint row,
			  gint column __attribute((unused)),
			  GdkEvent *event __attribute((unused)),
			  gpointer user_data __attribute((unused)))
{
  GtkWidget *names = lookup_widget(GTK_WIDGET(clist), "label51");
  GtkWidget *type = lookup_widget(GTK_WIDGET(clist), "label52");
  GtkWidget *spool_dir = lookup_widget(GTK_WIDGET(clist), "label53");
  GtkWidget *model = lookup_widget(GTK_WIDGET(clist), "label66");
  char *text = NULL;
  GPC_QUEUE_INFO *qinfo = NULL;
  char *qnames = NULL;
  
  if (row < 0)
    return;
  
  gtk_clist_get_text(clist, row, 0, &text);

  qinfo = gpc_queue_get_info(text);
  if (!qinfo)
    return;

  if (qinfo->filter_opts && qinfo->filter_opts->printer_model)
    gtk_label_set_text(GTK_LABEL(model), qinfo->filter_opts->printer_model);
  else
    gtk_label_set_text(GTK_LABEL(model), "N/A");
    
  qnames = gpc_queue_names(qinfo);
  gtk_label_set_text(GTK_LABEL(names), qnames);
  g_free(qnames);
  gtk_label_set_text(GTK_LABEL(type), prntype_names[qinfo->qtype]);
  gtk_label_set_text(GTK_LABEL(spool_dir), qinfo->spool_dir);
}


void on_clist2_unselect_row(GtkCList *clist,
			    gint row __attribute((unused)),
			    gint column __attribute((unused)),
			    GdkEvent *event __attribute((unused)),
			    gpointer user_data __attribute((unused)))
{
  GtkWidget *names = lookup_widget(GTK_WIDGET(clist), "label51");
  GtkWidget *type = lookup_widget(GTK_WIDGET(clist), "label52");
  GtkWidget *spool_dir = lookup_widget(GTK_WIDGET(clist), "label53");
  GtkWidget *model = lookup_widget(GTK_WIDGET(clist), "label66");

  gtk_label_set_text(GTK_LABEL(model), "N/A");
  gtk_label_set_text(GTK_LABEL(names), "N/A");
  gtk_label_set_text(GTK_LABEL(type), "N/A");
  gtk_label_set_text(GTK_LABEL(spool_dir), "N/A");
}


void
on_edit1_activate(GtkMenuItem *menuitem,
                  gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *wndMain = lookup_widget(GTK_WIDGET(menuitem), "wndMain");
  GtkWidget *clist = lookup_widget(wndMain, "clist2");
  GtkWidget *window1 = gtk_object_get_data(GTK_OBJECT(wndMain), AP);

  edt_printer_beg(window1, GTK_CLIST(clist));
}

void
on_dialog1_OK_clicked(GtkButton *button,
                      gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *dialog1 = lookup_widget(GTK_WIDGET(button), "dialog1");
  char *text = gtk_object_get_data(GTK_OBJECT(dialog1), "selected text");
  GtkWidget *clist = gtk_object_get_data(GTK_OBJECT(dialog1), "wndMain.clist");
  GtkWidget *entry = lookup_widget(GTK_WIDGET(dialog1), "txtAliasName");
  GPC_QUEUE_INFO *qinfo = gpc_queue_get_info(text);

  if (qinfo)
  {
    char *qname = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
    
    if (gpc_queue_get_info(qname))
    {
      error_have_printer_name(qname);
      return;
    }

    if (gpc_queue_alias(gpc_queue_get_list(NULL), qinfo, qname))
    {
      int row = 0;
      char *a[2];

      a[0] = qname;
      a[1] = NULL;
      
      row = gtk_clist_append(GTK_CLIST(clist), a);
      gtk_clist_select_row(GTK_CLIST(clist), row, 0);
    }
  }
  
  g_free(text);
  
  gtk_widget_destroy(dialog1);
}


void
on_dialog1_CANCEL_clicked(GtkButton *button,
                          gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *dialog1 = lookup_widget(GTK_WIDGET(button), "dialog1");

  gtk_widget_destroy(dialog1);
}

static void alias_printer_beg(GtkCList *clist)
{
  if (clist->selection)
  {
    int row = GPOINTER_TO_INT(clist->selection->data);
    char *text = NULL;
    
    gtk_clist_get_text(clist, row, 0, &text);

    if (row >= 0)
    {
      GtkWidget *dialog1 = create_dialog1();
      GtkWidget *label = lookup_widget(GTK_WIDGET(dialog1), "lblAliasDesc");
      GString *str = g_string_new("");

      g_string_sprintf(str, " Choose a name to be an alias for the printer "
                       "name -- %s --.\n", text);
      gtk_label_set_text(GTK_LABEL(label), str->str);
      g_string_free(str, TRUE);

      gtk_object_set_data(GTK_OBJECT(dialog1), "selected text", g_strdup(text));
     gtk_object_set_data(GTK_OBJECT(dialog1), "wndMain.clist",
                         (GtkWidget *)clist);
      
      gtk_widget_show(dialog1);
    }
  }
}

void
on_add_alias1_activate(GtkMenuItem *menuitem,
                       gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *clist = lookup_widget(GTK_WIDGET(menuitem), "clist2");

  alias_printer_beg(GTK_CLIST(clist));
}


void
on_btnAliasPrinter_clicked(GtkButton *button __attribute__ ((unused)),
                           gpointer user_data)
{
  GtkWidget *clist = user_data;
  
  alias_printer_beg(GTK_CLIST(clist));
}

void
on_restart_daemon1_activate(GtkMenuItem *menuitem __attribute__ ((unused)),
                            gpointer user_data __attribute__ ((unused)))
{ /* FIXME: should probably report error conditions better -- not easy */
  int val = system("/etc/init.d/lpd restart");

  if (val == -1)
  {
    gnome_error_dialog("Couldn't fork() to execute restart of lpd.\n");
  }
  else if (val)
  {
    int exit_code = WEXITSTATUS(val);
    GString *str = g_string_new("");

    g_string_sprintf(str, "The restart of lpd exited with exit code"
                     " -- %d --.\n", exit_code);
    gnome_error_dialog(str->str);
    g_string_free(str, TRUE);
  }
}

void
on_print_test_page_ascii1_activate(GtkMenuItem *menuitem,
                                   gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *clist = lookup_widget(GTK_WIDGET(menuitem), "clist2");
  GString *str = NULL;
  int row = -1;
  char *text = NULL;
  int val = -1;
  
  if (!GTK_CLIST(clist)->selection)
    return;
  
  row = GPOINTER_TO_INT(GTK_CLIST(clist)->selection->data);

  if (row < 0)
    return;
    
  gtk_clist_get_text(GTK_CLIST(clist), row, 0, &text);
  
  if (strchr(text, '"'))
  {
    gnome_warning_dialog("Doesn't work if you have a \" in the printer"
                         " name.\n");
    return;
  }

  str = g_string_new("");

  /* FIXME: this needs to be #define land so it can find the testpage */
  g_string_sprintf(str, "lpr \"-P%s\" %s/%s",
                   text, "/usr/share/printfilters", "testpage.asc");

  val = system(str->str);

  if (val == -1)
  {
    gnome_error_dialog("Couldn't fork() to print an ASCII testpage.\n");
  }
  else if (val)
  {
    int exit_code = WEXITSTATUS(val);

    g_string_sprintf(str, "The command for printing an ASCII test page"
                     " failed with exit code -- %d --.\n", exit_code);
    gnome_error_dialog(str->str);
  }
  
  g_string_free(str, TRUE);
}


void
on_print_test_page_ps1_activate(GtkMenuItem *menuitem,
                                gpointer user_data __attribute__ ((unused)))
{ /* FIXME: copy and paste from ASCII testpage ... */
  GtkWidget *clist = lookup_widget(GTK_WIDGET(menuitem), "clist2");
  GString *str = NULL;
  int row = -1;
  char *text = NULL;
  int val = -1;
  
  if (!GTK_CLIST(clist)->selection)
    return;
  
  row = GPOINTER_TO_INT(GTK_CLIST(clist)->selection->data);

  if (row < 0)
    return;
    
  gtk_clist_get_text(GTK_CLIST(clist), row, 0, &text);
  
  if (strchr(text, '"'))
  {
    gnome_warning_dialog("Doesn't work if you have a \" in the printer"
                         " name.\n");
    return;
  }

  str = g_string_new("");

  /* FIXME: this needs to be #define land so it can find the testpage */
  g_string_sprintf(str, "lpr \"-P%s\" %s/%s",
                   text, "/usr/share/printfilters", "testpage.ps");

  val = system(str->str);

  if (val == -1)
  {
    gnome_error_dialog("Couldn't fork() to print an PostScript testpage.\n");
  }
  else if (val)
  {
    int exit_code = WEXITSTATUS(val);

    g_string_sprintf(str, "The command for printing an PostScript test page"
                     " failed with exit code -- %d --.\n", exit_code);
    gnome_error_dialog(str->str);
  }
  
  g_string_free(str, TRUE);
}


void
on_clist1_select_row(GtkCList        *clist,
                     gint row __attribute__ ((unused)),
                     gint column __attribute__ ((unused)),
                     GdkEvent *event __attribute__ ((unused)),
                     gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *tmp = lookup_widget(GTK_WIDGET(clist), "btnNext");
  
  if (pagenum == WIZPG_AUTO_FOUND)
    gtk_widget_set_sensitive(GTK_WIDGET(tmp), TRUE);
}


void
on_clist1_unselect_row(GtkCList *clist,
                       gint row __attribute__ ((unused)),
                       gint column __attribute__ ((unused)),
                       GdkEvent *event __attribute__ ((unused)),
                       gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *tmp = lookup_widget(GTK_WIDGET(clist), "btnNext");
  
  if (pagenum == WIZPG_AUTO_FOUND)
    gtk_widget_set_sensitive(GTK_WIDGET(tmp), FALSE);
}


#define MV_GSTRING(ptr, str) do { \
  (ptr) = (str)->str; \
  g_string_free((str), FALSE); \
  (str) = g_string_new(""); \
  } while (FALSE)
      
void
on_btnFilter_clicked(GtkButton *button,
                     gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *diag2 = create_dialog2();
  GString *str = g_string_new("");
  GPC_PDINFO_RES **res = NULL;
  GPC_PDINFO_BITSPERPIXEL **bpp = NULL;
  unsigned int scan = 0;
  GtkWidget *tmp = NULL;
  GtkCList *clist = NULL;
  GtkWidget *window1 = lookup_widget(GTK_WIDGET(button), "window1");
  GPC_PDINFO *pdinfo = gtk_object_get_data(GTK_OBJECT(window1), "pdinfo entry");
  int done = FALSE;
  size_t len = 0;
  GPC_FILTER_SELECTION *old = NULL;
  int sel_row = 0;
  int row = 0;

  tmp = lookup_widget(diag2, "button5"); /* Apply */
  gtk_widget_hide(tmp);
  
  gtk_object_set_data(GTK_OBJECT(diag2), "window1", (gpointer)window1);
  old = gtk_object_get_data(GTK_OBJECT(window1), "device filter options");

  tmp = lookup_widget(diag2, "lblPrinterModel");
  gtk_label_set_text(GTK_LABEL(tmp), pdinfo->brief_desc);
  
  tmp = lookup_widget(diag2, "lblDriverDesc");
  gtk_label_set_text(GTK_LABEL(tmp), pdinfo->desc);

  res = pdinfo->res_list;
  bpp = pdinfo->bits_list;
  
  tmp = lookup_widget(diag2, "clistResolution");
  clist = GTK_CLIST(tmp);
  gtk_clist_freeze(clist);
  gtk_clist_clear(clist);

  if (old && old->res)
    len = strlen(old->res);
  
  sel_row = 0;
  done = FALSE;
  scan = 0;
  if (res)
    while (res[scan])
    {
      char *a[2];
      int row = 0;
      
      if (res[scan]->comment && strcmp(res[scan]->comment, "No comment."))
        g_string_sprintf(str, "%u%s%u (%s)", res[scan]->x, "x", res[scan]->y,
                         res[scan]->comment);
      else
        g_string_sprintf(str, "%u%s%u", res[scan]->x, "x", res[scan]->y);
      
      a[0] = str->str;
      a[1] = NULL;

      done = TRUE;
      row = gtk_clist_append(clist, a);

      if (old && old->res && !strncmp(old->res, str->str, len))
        sel_row = row;
      
      ++scan;
    }
  if (done)
  {
    gtk_widget_show(lookup_widget(diag2, "scrolledwindow4"));
    gtk_clist_select_row(clist, sel_row, 0);
  }
  else
    gtk_widget_hide(lookup_widget(diag2, "scrolledwindow4"));
  gtk_clist_thaw(clist);

  tmp = lookup_widget(diag2, "clistSize");
  clist = GTK_CLIST(tmp);
  gtk_clist_freeze(clist);
  gtk_clist_clear(clist);

  /* hard coded atm. ... should this come from somewhere ???
   * Eg. DJ 1220C has wide tray, most do not.
   * gs should give a way to find out which are the valid papersizes with a
   * certain driver... certainly the generic list of paper sizes in the gs
   * man page is bigger than this */
  if (TRUE)
  {
    char *a[2];

    sel_row = 0;
    a[1] = NULL;

    a[0] = (char *)"letter";       row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"legal";        row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"ledger";       row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"a3";           row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"a4";           row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"b5";           row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
    a[0] = (char *)"b4";           row = gtk_clist_append(clist, a);
    if (old && old->size && !strcmp(old->size, a[0]))
      sel_row = row;
  }
  gtk_clist_select_row(clist, sel_row, 0);
  gtk_clist_thaw(clist);
  
  tmp = lookup_widget(diag2, "clist5");
  clist = GTK_CLIST(tmp);
  gtk_clist_freeze(clist);
  gtk_clist_clear(clist);
  
  if (old && old->bpp)
    len = strlen(old->bpp);

  sel_row = 0;
  scan = 0;
  done = FALSE;
  if (bpp)
    while (bpp[scan])
    {
      char *a[2];
      
      if (res[scan]->comment && strcmp(res[scan]->comment, "No comment."))
        g_string_sprintf(str, "%s (%s)", bpp[scan]->bitsperpixel,
                         bpp[scan]->comment);
      else
        g_string_sprintf(str, "%s", bpp[scan]->bitsperpixel);
      
      a[0] = str->str;
      a[1] = NULL;

      done = TRUE;
      row = gtk_clist_append(clist, a);

      if (old && old->bpp && !strncmp(old->bpp, str->str, len))
        sel_row = row;

      ++scan;
    }
  if (done)
  {
    gtk_widget_show(lookup_widget(diag2, "frame5"));
    gtk_clist_select_row(clist, sel_row, 0);
  }
  else
    gtk_widget_hide(lookup_widget(diag2, "frame5"));
  gtk_clist_thaw(clist);

  /* Do text setup ... */
  if (old && old->margin_right_left)
  {
    tmp = lookup_widget(diag2, "entry2");
    gtk_entry_set_text(GTK_ENTRY(tmp), old->margin_right_left);
  }

  if (old && old->margin_top_bottom)
  {
    tmp = lookup_widget(diag2, "entry3");
    gtk_entry_set_text(GTK_ENTRY(tmp), old->margin_top_bottom);
  }

  if (old && old->pages_per_page)
  {
    tmp = lookup_widget(diag2, "combo1");
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(tmp)->entry), old->pages_per_page);
  }

  if (old && old->gs_opts)
  {
    tmp = lookup_widget(diag2, "entry1");
    gtk_entry_set_text(GTK_ENTRY(tmp), old->gs_opts);
  }

  /* Do check box setup */
  
  tmp = lookup_widget(window1, "lblSetDefsHdr");
  if (GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(tmp), CT)) ==
      GPC_QTYPE_LOCAL)
  { /* is this right, or was it the fast printing one ... ? */
    tmp = lookup_widget(diag2, "chkSendEOF");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
  }
  else
  {
    tmp = lookup_widget(diag2, "chkSendEOF");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), FALSE);
  }

  if (old && old->fix_stair_text)
  {
    tmp = lookup_widget(diag2, "chkFixStepping");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), TRUE);
  }
  else
  {
    tmp = lookup_widget(diag2, "chkFixStepping");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), FALSE);
  }
  
  gtk_widget_show(diag2);
  
  g_string_free(str, TRUE);
}
#undef MV_GSTRING

/* DUP text, can be NULL */
#define DUP_CLIST_TEXT(ptr, wid_name) do { \
  tmp = lookup_widget(diag2, wid_name); \
  g_assert(tmp); \
  text = clist_get_first_selected_text(GTK_CLIST(tmp)); \
  (ptr) = g_strdup(text); \
 } while (FALSE)

/* convert empty string to NULL */
#define DUP_ENTRY_TEXT(ptr, wid_name) do { \
  tmp = lookup_widget(diag2, wid_name); \
  g_assert(tmp); \
  (ptr) = gtk_editable_get_chars(GTK_EDITABLE(tmp), 0, -1); \
  if (!*(ptr)) { g_free(ptr); (ptr) = NULL; } \
 } while (FALSE)

#define DUP_TOG_CHOICE(bit, wid_name) do { \
  tmp = lookup_widget(diag2, wid_name); \
  g_assert(tmp); \
  (bit) = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp)); \
 } while (FALSE)

static void gpc__filter_apply(GtkWidget *diag2)
{
  GtkWidget *window1 = NULL;
  GtkWidget *tmp = NULL;
  GPC_FILTER_SELECTION *filter = g_malloc0(sizeof(GPC_FILTER_SELECTION));
  GPC_FILTER_SELECTION *old = NULL;
  char *text = NULL;
  GPC_PDINFO *pdinfo = NULL;
  
  DUP_CLIST_TEXT(filter->res, "clistResolution");
  DUP_CLIST_TEXT(filter->size, "clistSize");
  DUP_CLIST_TEXT(filter->bpp, "clist5");
  
  DUP_ENTRY_TEXT(filter->margin_right_left, "entry2");
  DUP_ENTRY_TEXT(filter->margin_top_bottom, "entry3");
  DUP_ENTRY_TEXT(filter->gs_opts, "entry1");

  tmp = lookup_widget(diag2, "combo1");
  filter->pages_per_page =
    gtk_editable_get_chars(GTK_EDITABLE(GTK_COMBO(tmp)->entry), 0, -1);
  
  DUP_TOG_CHOICE(filter->send_eof, "chkSendEOF");
  DUP_TOG_CHOICE(filter->fix_stair_text, "chkFixStepping");
  DUP_TOG_CHOICE(filter->fast_text, "chkFastText");

  window1 = gtk_object_get_data(GTK_OBJECT(diag2), "window1");
  old = gtk_object_get_data(GTK_OBJECT(window1), "device filter options");
  gpc__queue_free_filter_opts(old);
  gtk_object_set_data(GTK_OBJECT(window1), "device filter options",
                      (gpointer)filter);

  pdinfo = gtk_object_get_data(GTK_OBJECT(window1), "pdinfo entry");
  filter->printer_model = g_strdup(pdinfo->brief_desc);
  filter->gs_dev = g_strdup(pdinfo->gsdriver);
}
#undef DUP_CLIST_TEXT
#undef DUP_ENTRY_TEXT
#undef DUP_TOG_CHOICE


void
on_Filter_ok_clicked(GtkButton *button,
                     gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *diag2 = lookup_widget(GTK_WIDGET(button), "dialog2");

  gpc__filter_apply(diag2);
  gtk_widget_destroy(diag2);
}


void /* invis, don't need it */
on_Filter_apply_clicked(GtkButton *button,
                        gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *diag2 = lookup_widget(GTK_WIDGET(button), "dialog2");

  gpc__filter_apply(diag2);
}


void
on_Filter_cancel_clicked(GtkButton *button,
                         gpointer user_data __attribute__ ((unused)))
{
  GtkWidget *diag2 = lookup_widget(GTK_WIDGET(button), "dialog2");

  gtk_widget_destroy(diag2);
}

