/* Copyright (c) 2001 Mark Pruett, Ben Woodard, James Antill
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version 2
 * of the Licence, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRENTY; without even the implied warrenty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence in the COPYING file for more
 * details.
 *
 * You should have received a copy of the GNU Library General 
 * Public License along with the GNU C Library; see the file 
 * COPYING.LIB.  If not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <regex.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <gnome.h>

#include <gnome-xml/parser.h>

#include "gpc_pdinfo.h"
#include "gpc_impclient.h"

#define PDDB_PREFIX "PDDB"

// prototypes to remove warnings
// shouldn't these be static -ben 8/14/2001
int pdinfo_parse_xml_memory(char *url,char *manuf,xmlDocPtr * doc,
			    xmlNsPtr * ns,xmlNodePtr * cur);
int gpc_pdinfo_testmain (int argc, char **argv);
int gpc_pdinfo_init( int dbtype, char *url);
int gpc_pdinfo_get_dbtype(void);

/*================================================================
  Enums and variables
================================================================*/

static int global_dbtype = PDDB_LOCAL;
static char *global_url = NULL;
static char *global_manuf = NULL;

/* hacks to make it compile.... *sigh*/
#define url_to_char(x, y) (NULL)
#define url_to_charlist(x, y) (NULL)

/*-------------------------------------------------------------------
  Remove all leading and trailing whitespace ([\n\r\t ]) from str,
  in place.  Return pointer to chopped char string (str).
-------------------------------------------------------------------*/
static char * chop (char * str) {
  int front;
  int cur;

  if (str == NULL) return str;

  /* Find first non-blank character. */
  for (cur=0; strchr(" \r\n\t", str[cur]) && (str[cur] != '\0'); cur++); 

  /* Shift string to left. */
  for (front=0; str[cur] != '\0'; cur++) 
    str[front++] = str[cur];

  str[front] = '\0';

  /* Trim the right end of the string. */
  for (cur=strlen(str)-1; cur > 0; cur--)
    if (strchr (" \r\n\t", str[cur]))  
      str[cur] = '\0';
    else
      break;

  return str;
}

/*-------------------------------------------------------------------
  Replace the first occurrence of "search" found in string "str"
  with the string "replace".
-------------------------------------------------------------------*/
static char * strreplace (char *str, const char *search, const char *replace) {

  char *p = NULL;
  char *newstr = NULL;

  p = (char *) strstr (str, search);
  if (! p) return (char *) strdup (str);

  newstr = (char *) malloc (strlen (str) + strlen (replace) 
			    - strlen (search) + 10);
  strncpy (newstr, str, p-str);
  newstr[p-str] = 0;
  strcat (newstr, replace);
  p = p + strlen(search);
  if (p)
    strcat (newstr, p);

  return newstr;
}

/*===================================================================
  XML SUPPORT FUNCTIONS
===================================================================*/

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static char * xml_get_value_by_name (xmlAttr *cur, const char *name) {

  while (cur) {
    if (! strcmp (cur->name, name)) 
      return cur->val->content;
    cur = cur->next;
  }
  return NULL;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static char * pdinfo_xml_get_content (xmlNodePtr cur) {
  char *valstr = NULL;

  if (cur->childs->content)
    valstr = strdup (cur->childs->content);
  chop (valstr);

  return valstr;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static char * pdinfo_convert_escaped_chars (char *value) {

  char *tempstr1 = NULL;
  char *tempstr2 = NULL;

  if (! value) return NULL;

  tempstr1 = strreplace (value, "<", "&lt;");
  tempstr2 = strreplace (tempstr1, ">", "&gt;");
  free (tempstr1);
  tempstr1 = strreplace (tempstr2, "'", "&apos;");
  free (tempstr2);
  tempstr2 = strreplace (tempstr1, "\"", "&quot;");
  free (tempstr1);
  tempstr1 = strreplace (tempstr2, "&", "&amp;");
  free (tempstr2);

  return tempstr1;
}
  
/*===================================================================
  RESOLUTION (GPC_PDINFO_RES) SUPPORT FUNCTIONS.
===================================================================*/

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_free_res( GPC_PDINFO_RES **res) {
  if (res) {
    int idx;

    for(idx=0;res[idx];idx++) {
      free (res[idx]->comment);
      free (res[idx]);
    }
    /* Free last, null-terminated element. */
    free (res[idx]);

    /* Free the pointer list. */
    free (res);
  }
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_print_res ( FILE *fp, const char *format, 
			       GPC_PDINFO_RES *res, int convert) {
  char *oldfmt = NULL;
  char *newfmt = NULL;
  char numstr[10];
  char *convstr = NULL;

  oldfmt = (char *) malloc (strlen(format) + 1);
  if (! oldfmt) return;
  strcpy (oldfmt, format);
  
  snprintf (numstr, 9, "%d", res->x);
  numstr[9] = '\0';
  newfmt = strreplace (oldfmt, "%x%", numstr);
  free (oldfmt);
  oldfmt = newfmt;

  snprintf (numstr, 9, "%d", res->y);
  numstr[9] = '\0';
  newfmt = strreplace (oldfmt, "%y%", numstr);
  free (oldfmt);
  oldfmt = newfmt;

  if (convert) 
    convstr = pdinfo_convert_escaped_chars (res->comment);
  else
    convstr = res->comment;
  newfmt = strreplace (oldfmt, "%comment%", convstr);
  if (convert) free (convstr);

  free (oldfmt);
  
  fprintf (fp ? fp: stdout, newfmt);

}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static GPC_PDINFO_RES * pdinfo_create_res ( int x, int y, char * comment) {

  GPC_PDINFO_RES *res = NULL;

  res = (GPC_PDINFO_RES *) malloc (sizeof (GPC_PDINFO_RES));
  if (res) {
    res->x = x;
    res->y = y;
    res->comment = NULL;
    if (comment) {
      res->comment = (char *) malloc (strlen (comment) + 1);
      if (res->comment) 
	strcpy (res->comment, comment);
    }
  }
  return res;
}

static char **gpc_pdinfo_get_names( xmlNodePtr cur_res){
  int ridx;
  char **retval=malloc(sizeof(char*));
  for(ridx=0;cur_res;cur_res = cur_res->next){
    char *tmp = xml_get_value_by_name(cur_res->properties, "text");

    if (cur_res->type == XML_COMMENT_NODE)
      continue;
    
    if (!tmp)
    {
      g_assert_not_reached();
      continue;
    }
    
    retval[ridx++]=strdup(tmp);
    retval=(char**)realloc(retval,sizeof(char*)*(ridx+1));
  }
  retval=(char**)realloc(retval,sizeof(char*)*(ridx+1));
  if(retval)
    retval[ridx]=NULL;
  return retval;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static GPC_PDINFO_RES ** pdinfo_xml_get_resolution( xmlNodePtr cur_res) {
  int ridx;
  GPC_PDINFO_RES **res_list = NULL;

  for (ridx=0;cur_res;cur_res = cur_res->next) {
    int x = atoi (xml_get_value_by_name (cur_res->properties, "x"));
    int y = atoi (xml_get_value_by_name (cur_res->properties, "y"));
		      
    GPC_PDINFO_RES *res=
      pdinfo_create_res(x,y,xml_get_value_by_name(cur_res->properties, 
						  "comment"));

    /* Add item to resolution list. */
    res_list=
      (GPC_PDINFO_RES **)realloc(res_list,sizeof(GPC_PDINFO_RES*)*(ridx+1));
    if (res_list)
      res_list[ridx++] = res;
  }

  /* Null-terminate the resolution list. */
  res_list = (GPC_PDINFO_RES **) 
    realloc (res_list,sizeof(GPC_PDINFO_RES *) * (ridx+1));
  if (res_list)
    res_list[ridx] = NULL;

  return res_list;
}

/*===================================================================
  BITSPERPIXEL (GPC_PDINFO_BITSPERPIXEL) SUPPORT FUNCTIONS
===================================================================*/

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_free_bitsperpixel( GPC_PDINFO_BITSPERPIXEL **bits) {
  int idx = 0;

  if (bits) {
    while (bits[idx]) {
      free (bits[idx]->bitsperpixel);
      free (bits[idx]->comment);
      free (bits[idx]);
      idx++;
    }
    /* Free last, null-terminated element. */
    free (bits[idx]);

    /* Free the pointer list. */
    free (bits);
  }
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static GPC_PDINFO_BITSPERPIXEL *pdinfo_create_bitsperpixel(char *bitsperpixel,
							   char * comment) {

  GPC_PDINFO_BITSPERPIXEL *bits = NULL;

  bits = (GPC_PDINFO_BITSPERPIXEL *) malloc (sizeof (GPC_PDINFO_BITSPERPIXEL));
  if (bits) {
    bits->bitsperpixel = NULL;
    bits->comment = NULL;
    if (bitsperpixel) {
      bits->bitsperpixel = (char *) malloc (strlen (bitsperpixel) + 1);
      if (bits->bitsperpixel) 
	strcpy (bits->bitsperpixel, bitsperpixel);
    }
    if (comment) {
      bits->comment = (char *) malloc (strlen (comment) + 1);
      if (bits->comment) 
	strcpy (bits->comment, comment);
    }
  }
  return bits;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_print_bitsperpixel (FILE * fp, const char *format, 
				       GPC_PDINFO_BITSPERPIXEL *bits, 
				       int convert) {
  char *oldfmt = NULL;
  char *newfmt = NULL;
  char *convstr = NULL;

  oldfmt = (char *) malloc (strlen(format) + 1);
  if (! oldfmt) return;
  strcpy (oldfmt, format);
  
  if (convert)
    convstr = pdinfo_convert_escaped_chars (bits->bitsperpixel);
  else
    convstr = bits->bitsperpixel;
  newfmt = strreplace (oldfmt, "%bitsperpixel%", convstr);
  if (convert) free (convstr);
  free (oldfmt);
  
  oldfmt = newfmt;
  if (convert)
    convstr = pdinfo_convert_escaped_chars (bits->comment);
  else
    convstr = bits->comment;
  newfmt = strreplace (oldfmt, "%comment%", convstr);
  if (convert) free (convstr);
  free (oldfmt);
  
  fprintf (fp ? fp : stdout, newfmt);

}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static GPC_PDINFO_BITSPERPIXEL **pdinfo_xml_get_bitsperpixel(xmlNodePtr cur_bits) {
  int bidx;
  GPC_PDINFO_BITSPERPIXEL *bits = NULL;
  GPC_PDINFO_BITSPERPIXEL **bits_list = NULL;
  
  bidx = 0;
  while (cur_bits) {
    bits = pdinfo_create_bitsperpixel 
      (xml_get_value_by_name (cur_bits->properties, "bits"),
       xml_get_value_by_name (cur_bits->properties, "comment"));

    /* Add item to bits list. */
    bits_list = (GPC_PDINFO_BITSPERPIXEL **) 
      realloc (bits_list, sizeof (GPC_PDINFO_BITSPERPIXEL *) * (bidx+1));
    if (bits_list) {
      bits_list[bidx++] = bits;
    }

    cur_bits = cur_bits->next;
  }

  /* Null-terminate bits list. */
  bits_list = (GPC_PDINFO_BITSPERPIXEL **) 
    realloc (bits_list, sizeof (GPC_PDINFO_BITSPERPIXEL *) * (bidx+1));
  bits_list[bidx] = NULL;
	  
  return bits_list;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_init (GPC_PDINFO * pdinfo) {

  pdinfo->printer_name = NULL;
  pdinfo->printer_manuf = NULL;
  pdinfo->gsdriver = NULL;
  pdinfo->brief_desc = NULL;
  pdinfo->desc = NULL;
  pdinfo->ppdfile = NULL;

  pdinfo->res_list = NULL;
  pdinfo->bits_list = NULL;
}

/*----------------------------------------------------------------
  Do preliminary sanity checking on XML file, and return pointer
  to the XML tree and namespace if okay, or NULLs otherwise.

  Returns 0 on success, non-zero on failure.
----------------------------------------------------------------*/
static int pdinfo_parse_xml_file(const char *filename,xmlDocPtr *doc, 
				 xmlNsPtr *ns, xmlNodePtr *cur)
{
    /* Check for existence of file. */
    if (! g_file_exists (filename)) return 1;

    /* Build an XML tree from a the file */
    *doc = xmlParseFile(filename);
    if (*doc == NULL) return 1;

    /* Check the document is of the right kind */
    *cur = (*doc)->root;
    if (*cur == NULL) {
        fprintf(stderr,"empty document\n");
	xmlFreeDoc(*doc);
	return 1;
    }

    *ns = xmlSearchNsByHref(*doc, *cur, "urn:linux-printerdb");
    if ((*ns == NULL) || (strcmp((*cur)->name, "printerdb"))) {
        fprintf(stderr,
	        "Document of the wrong type\n");
	xmlFreeDoc (*doc);
	return 1;
    }

    return 0;
}

/*----------------------------------------------------------------
  Do preliminary sanity checking on remote XML information, 
  and return pointer
  to the XML tree and namespace if okay, or NULLs otherwise.

  Returns 0 on success, non-zero on failure.
----------------------------------------------------------------*/
int pdinfo_parse_xml_memory(char *url __attribute((unused)),
			    char *manuf,xmlDocPtr * doc,
			    xmlNsPtr * ns,xmlNodePtr * cur) {
  char * buffer = NULL;
  char *urlstr = NULL;
  int buflen = 0;

  /* Get data from remote server. */
  urlstr = (char *) malloc (strlen (global_url) + 100);
  sprintf (urlstr, "%s?action=printer-entry-data&manufacturer=%s", global_url, manuf);
  buffer = url_to_char (urlstr, &buflen);
  free (urlstr);

  printf ("buffer [%s]\n", buffer);

  /* Check for existence of file. */
  if (! buffer) return 1;

  /* Build an XML tree from a the file */
  *doc = xmlParseMemory(buffer, buflen);
  if (*doc == NULL) return 1;

    /* Check the document is of the right kind */
  *cur = (*doc)->root;
  if (*cur == NULL) {
    fprintf(stderr,"empty document\n");
    xmlFreeDoc(*doc);
    return 1;
  }

  *ns = xmlSearchNsByHref(*doc, *cur, "urn:linux-printerdb");
  if ((*ns == NULL) || (strcmp((*cur)->name, "printerdb"))) {
    fprintf(stderr,
	    "Document of the wrong type\n");
    xmlFreeDoc (*doc);
    return 1;
  }

  printf ("Done.\n");
  return 0;
}

/*===================================================================
  PDDB API FUNCTIONS
===================================================================*/

/*-------------------------------------------------------------------
  Used to qsort() the pdinfo list.
-------------------------------------------------------------------*/
static int pdinfo_compar (const void *el0, const void *el1) {

  GPC_PDINFO * p0 = * ((GPC_PDINFO **) el0);
  GPC_PDINFO * p1 = * ((GPC_PDINFO **) el1);

  return (strcmp (p0->printer_name[0], p1->printer_name[0]));
}

/*-------------------------------------------------------------------
  Insert an item into the internal PDDB list.

  Given a list and a new item, insert the new item into the list
  in sorted (by printer_name) order.

  Returns the new list.
-------------------------------------------------------------------*/
GPC_PDINFO ** gpc_pdinfo_insert ( GPC_PDINFO ** list, GPC_PDINFO *item) {
  int idx = 0;

  /* Don't insert null items */
  if (! item)
    return list;

  /* Is this a new list? */
  if (list == NULL) {
    list = (GPC_PDINFO **) realloc (list, sizeof (GPC_PDINFO *) * (idx+1));
    list[idx] = NULL;
  }

  /* Find end of list */
  while (list[idx]) 
    idx++;

  /* Insert item */
  if (list[idx] == NULL)
    list[idx++] = item;

  /* Null-terminate the list. */
  list = (GPC_PDINFO **) realloc (list, sizeof (GPC_PDINFO *) * (idx+1));
  list[idx] = NULL;

  /* Sort the list. */
  if (idx > 1)
    qsort (list, idx, sizeof (GPC_PDINFO *), pdinfo_compar);

  return list;
}

/*----------------------------------------------------------------
  Find a printer in the printer list.

  Given the list and the name of a printer, search the list and 
  find an item whose printer_name field matches.

  Returns a pointer to the matching list item if found, or 
  NULL otherwise.
-------------------------------------------------------------------*/
GPC_PDINFO * gpc_pdinfo_find(GPC_PDINFO **list,const char *printer_name){
  int idx = 0;

  while (list[idx]) {
    char **curname;
    for(curname=list[idx]->printer_name;*curname!=NULL;curname++)
      if (!strcmp (printer_name, *curname)) 
	return list[idx];
    idx++;
  }

  return NULL;
}

/*----------------------------------------------------------------
  Find a printer in the printer list.

  Given the list and the name of a printer, search the list and 
  find an item whose printer_name field matches.

  Returns a pointer to the matching list item if found, or 
  NULL otherwise.
-------------------------------------------------------------------*/
GPC_PDINFO *gpc_pdinfo_description_find(GPC_PDINFO **list,
                                        const char *printer_desc)
{
  int idx = 0;

  while (list[idx])
  {
    if (!strcmp (printer_desc, list[idx]->brief_desc)) 
      return list[idx];
    
    idx++;
  }

  return NULL;
}

/*-------------------------------------------------------------------
  Free a previously-allocated GPC_PDINFO list.

  Returns NULL.
-------------------------------------------------------------------*/
GPC_PDINFO ** gpc_pdinfo_free_list( GPC_PDINFO **list) {
  int idx = 0;

  if (! list) return NULL;

  idx = 0;
  while (list[idx]) {
    char **curname;
    /* Free sub-entries */
    for(curname=list[idx]->printer_name;*curname!=NULL;curname++)
      free(curname);
    free (list[idx]->printer_name);
    free (list[idx]->printer_manuf);
    free (list[idx]->gsdriver);
    free (list[idx]->brief_desc);
    free (list[idx]->desc);
    free (list[idx]->ppdfile);

    /* Free resolutions */
    pdinfo_free_res (list[idx]->res_list);

    /* Free bitsperpixel */
    pdinfo_free_bitsperpixel (list[idx]->bits_list);

    /* Free list node itself */
    free (list[idx]);

    idx++;
  }

  /* Delete list itself. */
  free (list);

  return NULL;
}

/*----------------------------------------------------------------
  Parse the xml file containing pddb information.

  Given a file name of a local PDDB file, this routine parses the
  file and produces a list of GPC_PDINFO structures containing
  the file's contents.

  Returns a null-terminated array of PDDB records, or NULL if 
  an error occurs.
----------------------------------------------------------------*/
GPC_PDINFO ** gpc_pdinfo_get_list(const char *filename,
				  char **url_list __attribute((unused))) {
/* filename -- the name of the PDDB (XML formatted) file */
/* url_list -- reserved for future use. */

    xmlDocPtr doc = NULL;
    xmlNsPtr ns = NULL;
    xmlNodePtr cur_entry;
    xmlNodePtr cur = NULL;

    int idx;		/* Index for pdinfo list. */
    GPC_PDINFO *pdinfo = NULL;
    GPC_PDINFO **list = NULL;

    if (global_dbtype == PDDB_REMOTE) {
      pdinfo_parse_xml_memory (global_url, global_manuf, &doc, &ns, &cur);
    }
    else {
      if (pdinfo_parse_xml_file (filename, &doc, &ns, &cur)) {
	fprintf (stderr, "Unrevoverable problem with file '%s'.\n",
		 filename);
	exit (1);
      }
    }

    /* Now, walk the tree. */

    /* First level we expect "entry" */
    cur = cur->childs;

    cur_entry = cur;
    idx = 0;

    while (cur_entry) {
      if ((strcmp(cur_entry->name, "entry")) || (cur_entry->ns != ns)) {
        fprintf(stderr,
		"document of the wrong type, Tabs expected, found (%s)\n",
		cur_entry->name);
	xmlFreeDoc (doc);
	return NULL;
      }

      /* Second level are sub-fields of "entry": description, driver, etc. */
      cur = cur_entry->childs;
      /* Create a new struct to hold this entry. */
      pdinfo = (GPC_PDINFO *) malloc (sizeof (GPC_PDINFO));
      if (! pdinfo) {
	fprintf (stderr, "Can't malloc GPC_PDINFO structure.\n");
	exit (1);
      }
      pdinfo_init (pdinfo);

      while (cur != NULL) {
        if ((!strcmp(cur->name, "entrynames")) && (cur->ns == ns)) {
	  pdinfo->printer_name = gpc_pdinfo_get_names (cur->childs);
	}else if ((!strcmp(cur->name, "manufacturer")) && (cur->ns == ns)) {
	  pdinfo->printer_manuf = pdinfo_xml_get_content (cur);
	}else if ((!strcmp(cur->name, "description")) && (cur->ns == ns)) {
	  pdinfo->brief_desc = pdinfo_xml_get_content (cur);
	}else if ((!strcmp(cur->name, "about")) && (cur->ns == ns)) {
	  pdinfo->desc = pdinfo_xml_get_content (cur);
	}else if ((!strcmp(cur->name, "driver")) && (cur->ns == ns)) {
	  pdinfo->gsdriver = pdinfo_xml_get_content (cur);
	}else if ((!strcmp(cur->name, "ppdfile")) && (cur->ns == ns)) {
	  pdinfo->ppdfile = pdinfo_xml_get_content (cur);
	}else if ((!strcmp(cur->name, "bitslist")) && (cur->ns == ns)) { 
	  /* BITSPERPIXEL */
	  pdinfo->bits_list = pdinfo_xml_get_bitsperpixel (cur->childs);
	}else if ((!strcmp(cur->name, "resolutions")) && (cur->ns == ns)) { 
	  /* RESOLUTION */
	  pdinfo->res_list = pdinfo_xml_get_resolution (cur->childs);
	}else {
	  printf ("Unknown: %s\n", cur->name);
	}
	cur = cur->next;
      }

      /* Add this node to the list */

      list = gpc_pdinfo_insert (list, pdinfo);

      cur_entry = cur_entry->next;
    }

    xmlFreeDoc (doc);
    return list;
}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static void pdinfo_print_list (FILE *fp, GPC_PDINFO **list) {
  int idx = 0;
  int bidx;
  int ridx;

  if (list == NULL) return;

  fp = fp ? fp : stdout;

  while (list[idx]) {
    char **curname;
    fprintf (fp, "====================================================\n");
    fprintf (fp, "Printer Names:\n");
    for(curname=list[idx]->printer_name;*curname!=NULL;curname++)
      fprintf(fp,"\t%s\n",*curname);
    fprintf (fp, "\tPPD File: %s\n", list[idx]->ppdfile ? list[idx]->ppdfile : "None");

    bidx = 0;
    ridx = 0;

    if (list[idx]->bits_list)
      while (list[idx]->bits_list[bidx])
	pdinfo_print_bitsperpixel 
	  (fp, "bitsperpixel = %bitsperpixel%, comment = '%comment%'\n",
	   list[idx]->bits_list[bidx++], 0);

    if (list[idx]->res_list)
      while (list[idx]->res_list[ridx])
	pdinfo_print_res 
	  (fp, "resolution (%x%, %y%)\n",
	   list[idx]->res_list[ridx++], 0);

      
    idx++;
  }
}

/*----------------------------------------------------------------
  Print a simple PDDB sub-entry in XML format.
----------------------------------------------------------------*/
static void pdinfo_xml_print_item ( FILE *fp, const char *name, 
				    char *value){
  char *converted_val = NULL;

  if (value) {
    converted_val = pdinfo_convert_escaped_chars (value);

    fprintf (fp, "  <%s:%s>\n    %s\n  </%s:%s>\n",
	     PDDB_PREFIX, name, converted_val, PDDB_PREFIX, name);

    free(converted_val);
  }
}

/*----------------------------------------------------------------
  Save the GPC_PDINFO list to an local PDDB XML file.

  Returns 0 on success, non-zero otherwise;
-------------------------------------------------------------------*/
int gpc_pdinfo_save_local( GPC_PDINFO **list, const char *filename)
{
  int idx = 0;
  int bidx;
  int ridx;
  FILE *fp = NULL;
  const char *bits_tmpl = ("    <%s%s:bitsperpixel bits=\"%%bitsperpixel%%\" "
                           "comment=\"%%comment%%\" />\n");
  const char *res_tmpl = ("    <%s%s:resolution x=\"%%x%%\" y=\"%%y%%\" "
                          "comment=\"%%comment%%\" />\n");
  char *bits_fmt = NULL;
  char *res_fmt = NULL;

  bits_fmt = (char *) malloc (strlen (bits_tmpl) + strlen (PDDB_PREFIX) + 2); 
  sprintf (bits_fmt, bits_tmpl, "", PDDB_PREFIX); 

  res_fmt = (char *) malloc (strlen (res_tmpl) + strlen (PDDB_PREFIX) + 1); 
  sprintf (res_fmt, res_tmpl, "", PDDB_PREFIX);

  /* open the file. */
  if (filename) 
    fp = (FILE *) fopen (filename, "w");
  else
    fp = stdout;

  fprintf (fp, "<%s:printerdb xmlns:%s=\"urn:linux-printerdb\">\n", 
	   PDDB_PREFIX, PDDB_PREFIX);

  while (list[idx]) {
    char **curname;

    fprintf (fp, "\n<%s:entry>\n", PDDB_PREFIX);

    for(curname=list[idx]->printer_name;*curname!=NULL;curname++)
      pdinfo_xml_print_item (fp, "entryname", *curname); 
    pdinfo_xml_print_item (fp, "manufacturer", list[idx]->printer_manuf); 
    pdinfo_xml_print_item (fp, "description", list[idx]->brief_desc); 
    pdinfo_xml_print_item (fp, "driver", list[idx]->gsdriver); 
    pdinfo_xml_print_item (fp, "about", list[idx]->desc); 
    pdinfo_xml_print_item (fp, "ppdfile", list[idx]->ppdfile); 

    bidx = 0;
    ridx = 0;

    if (list[idx]->bits_list) {
      fprintf (fp, "  <%s:bitslist>\n", PDDB_PREFIX);
      while (list[idx]->bits_list[bidx])
	pdinfo_print_bitsperpixel 
	  (fp, bits_fmt,
	   list[idx]->bits_list[bidx++], 1);
      fprintf (fp, "  </%s:bitslist>\n", PDDB_PREFIX);
    }

    if (list[idx]->res_list) {
      fprintf (fp, "  <%s:resolutions>\n", PDDB_PREFIX);
      while (list[idx]->res_list[ridx])
	pdinfo_print_res 
	  (fp, res_fmt,
	   list[idx]->res_list[ridx++], 1);

      fprintf (fp, "  </%s:resolutions>\n", PDDB_PREFIX);
    }
      
    fprintf (fp, "</%s:entry>\n", PDDB_PREFIX);

    idx++;
  }

  fprintf (fp, "\n</%s:printerdb>\n", PDDB_PREFIX);

  /* Close the file and clean up. */
  fclose (fp);

  free (bits_fmt);
  free (res_fmt);

  return 0;

}

/*-------------------------------------------------------------------
-------------------------------------------------------------------*/
static char ** charlist_append (char **list, char *item, int *num_items) {
  int idx = 0;

  /* Check if item is already in list. */
  if (list)
    while (list[idx]) {
      if (! strcmp (list[idx], item))
	return list;
      idx++;
    }
  else {
    list = (char **) malloc (sizeof (char *));
    list[idx] = NULL;
  }

  /* Add to end of list (item replaces NULL node at end of list). */
  list[idx] = (char *) strdup (item);
  idx++;

  list = (char **) realloc (list, sizeof (char *) * (idx+1));
  list[idx] = NULL;

  *num_items = idx;

  return list;
}  


/*-------------------------------------------------------------------
  Used by qsort in gpc_pdinfo_get_manuf_names();
-------------------------------------------------------------------*/
static int charlist_compar (const void *el0, const void *el1) {

  char * p0 = * ((char **) el0);
  char * p1 = * ((char **) el1);

  return (strcmp (p0, p1));
}

/*-------------------------------------------------------------------
  Frees a list of manufacturer's names

  Free a list of manufacturer's names previously allocated by
  gpc_pdinfo_get_manuf_names().
-------------------------------------------------------------------*/
void gpc_pdinfo_free_manuf_names (char **list) {

  int idx = 0;
  if (list) {
    if (global_dbtype == PDDB_LOCAL) {
      while (list[idx]) {
	free (list[idx]);
	idx++;
      }
      /* Free null-terminated (last) list item. */
      free (list[idx]);
    }

    /* Free the list of pointers itself. */
    free (list);
  }
}

/*-------------------------------------------------------------------
  Get a list of unique manufacturers names.

  Given a list of PDDB records, build a list of unique manufacturers
  names, extracted from the printer_manuf fields.

  Returns a sorted, null-terminated list of manufacturer's names.
-------------------------------------------------------------------*/
char ** gpc_pdinfo_get_manuf_names ( GPC_PDINFO **list) {

  int idx = 0;
  char **manuf_list = NULL;
  int num_items = 0;
  char *urlstr = NULL;

  if (global_dbtype == PDDB_REMOTE) {
    urlstr = (char *) malloc (strlen (global_url) + 100);
    sprintf (urlstr, "%s?action=printer-manuf-info&headers=none", global_url);
    
    manuf_list = url_to_charlist (urlstr, &num_items);
    manuf_list = (char **) realloc (manuf_list, sizeof (char *) * (num_items + 1));
    manuf_list[num_items] = NULL;
    
  }
  else if (global_dbtype == PDDB_LOCAL) {
    while (list[idx]) {
      manuf_list = charlist_append (manuf_list, 
				    list[idx]->printer_manuf, 
				    &num_items);
      idx++;
    }
    qsort (manuf_list, num_items-1, sizeof (char *), charlist_compar);
  }

  return manuf_list;
}

/*----------------------------------------------------------------
  Free a list of pointers to PDDB records for a specific 
  manufacturer.  This routine doesn't delete the records 
  themselves, as they are still part of the complete list of PDDB
  records allocated with gpc_pdinfo_get_list().
----------------------------------------------------------------*/
void gpc_pdinfo_free_manuf_list(GPC_PDINFO **manlist) {
  /* manlist -- list of PDDB records for one manufacturer*/
  free (manlist);
}

/*----------------------------------------------------------------
  Get a list of PDDB records for a single manufacturer.

  Given the list of all PDDB records and the name of a target 
  manufacturer, return a list of pointers to matching records
  within the list.
----------------------------------------------------------------*/
GPC_PDINFO ** gpc_pdinfo_get_manuf_list ( GPC_PDINFO **list, char *manuf) {
  int idx = 0;
  int midx = 0;
  GPC_PDINFO **manlist = NULL;

  if (! list) return NULL;

  while (list[idx]) {
    if (! strcmp (manuf, list[idx]->printer_manuf)) {
      manlist = (GPC_PDINFO **) 
	realloc (manlist, sizeof (GPC_PDINFO *) * (midx+1));
      manlist[midx] = list[idx];
      midx++;
    }

    idx++;
  }
  if (manlist) {
    manlist = (GPC_PDINFO **) 
      realloc (manlist, sizeof (GPC_PDINFO *) * (midx+1));
    manlist[midx] = NULL;
  }

  return manlist;

} 

/*----------------------------------------------------------------
----------------------------------------------------------------*/
int gpc_pdinfo_testmain (int argc __attribute((unused)), 
			 char **argv __attribute((unused))) {
  GPC_PDINFO **list = NULL;
  GPC_PDINFO *item = NULL;
  char **manuf_list = NULL;
  int idx = 0;

  list = gpc_pdinfo_get_list ("printerdb.xml", NULL);

  pdinfo_print_list (stdout, list);

  item = gpc_pdinfo_find (list, "ALPSMD5K");
  printf ("!!!!!===============================================\n");
  printf ("Printer Name: %s\n", item->printer_name[0]);
  printf ("\tPPD File: %s\n", item->ppdfile ? item->ppdfile : "None");


  gpc_pdinfo_save_local (list, "/tmp/gpctest.xml");

  printf ("num items = %d\n", sizeof ((char **) list));


  /*--------------------------------------------------------------
    Test manufacturers list routines
  --------------------------------------------------------------*/
  manuf_list = gpc_pdinfo_get_manuf_names (list);

  idx = 0;
  while (manuf_list[idx]) {
    printf ("manuf = [%s]\n", manuf_list[idx]);
    idx++;
  }

  gpc_pdinfo_free_manuf_names (manuf_list);

  list = gpc_pdinfo_free_list (list);
  pdinfo_print_list (stdout, list);

  return 0;

}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
int gpc_pdinfo_init( int dbtype, char *url) {

  global_dbtype = dbtype;
  if (global_url) free (global_url);

  global_url = (char *) malloc (strlen (url) + 1);
  if (global_url) 
    strcpy (global_url, url);

  return 1;
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
int gpc_pdinfo_set_manufacturer( char * manuf) {

  if (global_manuf) free (global_manuf);
  global_manuf = (char *) strdup (manuf);

  return global_manuf ? 1 : 0;
}

/*----------------------------------------------------------------
----------------------------------------------------------------*/
int gpc_pdinfo_get_dbtype() {
  return global_dbtype;
}
